/*
    reloader.cpp - main file for IDA loader demonstration 

    This file is part of the reloader.
    Reloader is free software; you can redistribute it and/or
    modify it.

    This software uses parts of GNU libiberty library covered by 
    GPL version 2 license. 

*/

#include <string.h>
#include <stdio.h>

#include "reloader.h"

// IDA includes
#include <ida.hpp>
#include <fpro.h>
#include <idp.hpp>
#include <loader.hpp>
#include <name.hpp>
#include <bytes.hpp>
#include <offset.hpp>
#include <segment.hpp>
#include <srarea.hpp>
#include <fixup.hpp>
#include <entry.hpp>
#include <auto.hpp>
#include <diskio.hpp>
#include <kernwin.hpp>
#include <typeinf.hpp>

tid_t 	dos_header_struct;
tid_t 	nt_header_struct;
tid_t 	opt_header_struct;
tid_t 	section_header_struct;

// ------------------------ Global data ------------------------ 

void add_types()
{
	add_til( "reexe.til");
	dos_header_struct = til2idb(-1, "dos_header");
	nt_header_struct = til2idb(-1, "nt_header");
	opt_header_struct = til2idb(-1, "optional_header");
	section_header_struct = til2idb(-1, "section_header");
}

/*
 *   Check if input file can be a newton MP image file. If it passes return 
 *   and fill in the formatname otherwise return 0
 */
int accept_file(linput_t *li, char fileformatname[MAX_FILE_FORMAT_NAME], int n) {
    uint32 magic;
	if( n || lread4bytes( li, &magic, false) ) return 0;
    if (magic != RE_MAGIC ) {
        return 0;
    }

    // first get the lenght of the file
    qsnprintf(fileformatname, MAX_FILE_FORMAT_NAME, "REVERSE Executable" );

    return ACCEPT_FIRST | 1;
}

//
// Load the recognised file into the database.
//
void load_file(linput_t *li, ushort /*neflag*/, const char * /*fileformatname*/) 
{
	struct dos_header dhdr;
	uint32 pos=0;
	
	add_types();

	lread(li, &dhdr, sizeof( struct dos_header ) );
	mem2base( &dhdr, pos, pos+ sizeof( struct dos_header ), pos );

	if( !add_segm(0, pos, pos+sizeof( struct dos_header ), ".doshdr", "DATA" ) ) {
		loader_failure();
	}
	segment_t *dos_seg = getseg( 0 );
	set_segm_addressing(dos_seg, 0 );

	doStruct(pos, sizeof( struct dos_header ), dos_header_struct );
	
	pos += sizeof( struct dos_header );

	file2base( li, pos, pos, dhdr.e_lfanew, FILEREG_PATCHABLE );
	if( !add_segm(0, pos, dhdr.e_lfanew, ".dosprg", "DATA" ) ) {
		loader_failure();
	}
	segment_t *dos_prg = getseg( pos );
	set_segm_addressing(dos_prg, 0 );
	pos = dhdr.e_lfanew;

	// NT header
	struct nt_header nthdr; 
	lread(li, &nthdr, sizeof( struct nt_header ) );
	mem2base( &nthdr, pos, pos+ sizeof( struct nt_header ), pos );

	if( !add_segm(0, pos, pos+sizeof( struct nt_header ), ".nthdr", "DATA" ) ) {
		loader_failure();
	}
	segment_t *nt_seg = getseg( pos );
	set_segm_addressing(nt_seg, 1 );
	doStruct(pos, sizeof( struct nt_header ), nt_header_struct );

	pos+=sizeof( struct nt_header );


	// Optional header
	struct optional_header opthdr; 
	uint32 opt_header_pos = pos;
	lread(li, &opthdr, sizeof( struct optional_header ) );
	mem2base( &opthdr, pos, pos+ sizeof( struct optional_header ), pos );

	if( !add_segm(0, pos, pos+sizeof( struct optional_header ), ".opthdr", "DATA" ) ) {
		loader_failure();
	}
	segment_t *opt_seg = getseg( pos );
	set_segm_addressing(opt_seg, 1 );
	doStruct(pos, sizeof( struct optional_header), opt_header_struct );

	pos+=sizeof( struct optional_header );

	// HEADER DATA 
	file2base( li, pos, pos, opt_header_pos + nthdr.SizeOfOptionalHeader, FILEREG_PATCHABLE );
	if( !add_segm(0, pos, opt_header_pos + nthdr.SizeOfOptionalHeader, ".img_dat_dir", "DATA" ) ) {
		loader_failure();
	}
	segment_t *s = getseg( pos );
	set_segm_addressing(s, 0 );
	pos = opt_header_pos + nthdr.SizeOfOptionalHeader;

	uint32 sec_pos = pos;
	section_header* pSec=NULL;
	if( nthdr.NumberOfSection )
	{
		pSec = (section_header*) malloc( nthdr.NumberOfSection * sizeof(section_header) );
		lread(li, pSec, nthdr.NumberOfSection * sizeof( struct section_header ) );
		mem2base( pSec, pos, pos + nthdr.NumberOfSection * sizeof( struct section_header ), pos );
	
		if( !add_segm(0, pos,  pos + nthdr.NumberOfSection * sizeof(section_header), ".section_hrd", "DATA" ) ) {
			loader_failure();
		}
		segment_t *s = getseg( pos );
		set_segm_addressing(s, 1);
		for(int i=0; i< nthdr.NumberOfSection; i++)
		{
			doStruct(pos, sizeof( struct section_header), section_header_struct );
			pos+=sizeof( struct section_header);
		}
	}

	FILE *fi = qlfile(li);
	ulong file_length = efilelength(fi);
	file2base( li, pos, pos, file_length, FILEREG_PATCHABLE );
	uint32 sp=sec_pos;
	for(int i=0; i< nthdr.NumberOfSection; i++)
	{
		char* strType;
		if( strcmp( pSec[i].name, ".text") == 0 ) strType = "CODE";
		else strType = "DATA";
		if( !add_segm(0, pSec[i].PointereToRawData,  pSec[i].PointereToRawData+pSec[i].SizeOfRawData, pSec[i].name, strType ) ) {
			loader_failure();
		}
		segment_t *opt_seg = getseg( pSec[i].PointereToRawData );
		set_segm_addressing(opt_seg, 1 );
	}

	for(int i=nthdr.NumberOfSection-1; i>= 0; i--)
	{
		segment_t *s = getseg( pSec[i].PointereToRawData );
		move_segm( s, pSec[i].VirtualAddress, 0 );
		//set_segm_addressing(s, 1 );
	}

	rebase_program( opthdr.BaseOfImage, 0 );
	
	add_entry(opthdr.AddressOfEntryPoint+opthdr.BaseOfImage, opthdr.AddressOfEntryPoint+opthdr.BaseOfImage,"_start", true );
	create_filename_cmt();
}

int idaapi save_file( FILE* fp, const char* fileformatname )
{
	segment_t *s = getnseg(0);
	if( s )
	{
		base2file(fp, sizeof( struct dos_header ), s->startEA, s->endEA );
	}

	return 1;
}

//
// Loader Module Descriptor Blocks
//
extern "C" loader_t LDSC = {
  IDP_INTERFACE_VERSION,
  0, /* no loader flags */
  accept_file,
  load_file,
  save_file,
  NULL,
  NULL,
};

